
import argparse
import collections
import csv
import itertools
import json
import os.path
import subprocess
import sys

from decontamlib.version import __version__
from decontamlib.tools import FilteringTool


def get_config(user_config_file, organism):
    config = {
        "method": "bwa",
        "bowtie2_fp": "bowtie2",
        "bwa_fp":"bwa",
        "num_threads":8
    }

    if user_config_file is None:
        if organism == "host":
            default_user_config_fp = os.path.expanduser("~/.decontam_human.json")
        elif organism == "phix":
            default_user_config_fp = os.path.expanduser("~/.decontam_phix.json")
        if os.path.exists(default_user_config_fp):
            user_config_file = open(default_user_config_fp)

    if user_config_file is not None:
        user_config = json.load(user_config_file)
        config.update(user_config)
    return config


def human_filter_main(argv=None):
    p = argparse.ArgumentParser()

    # Input
    p.add_argument(
        "--forward-reads", required=True,
        type=argparse.FileType("r"),
        help="FASTQ file of forward reads")
    p.add_argument(
        "--reverse-reads", required=True,
        type=argparse.FileType("r"),
        help="FASTQ file of reverse reads")
    p.add_argument(
        "--config-file",
        type=argparse.FileType("r"),
        help="JSON configuration file")
    p.add_argument(
        "--organism", required=True,
        help="reference organism to filter from")
    p.add_argument(
        "--pct", required=False,
        type=float, default=0.5,
        help="Percent identity Default: 0.5")
    p.add_argument(
        "--frac", required=False,
        type=float, default=0.6,
        help="Fraction of alignment length Default: 0.6")
   
    # Output
    p.add_argument(
        "--summary-file", required=True,
        type=argparse.FileType("w"),
        help="long table of results.")
    p.add_argument(
        "--output-dir", required=True,
        help="Path to output directory")
    args = p.parse_args(argv)

    config = get_config(args.config_file, args.organism)

    fwd_fp = args.forward_reads.name
    rev_fp = args.reverse_reads.name
    args.forward_reads.close()
    args.reverse_reads.close()

    tool = FilteringTool(config)
    if not tool.index_exists():
        tool.make_index()

    if not os.path.exists(args.output_dir):
        os.mkdir(args.output_dir)

    summary_data = tool.decontaminate(fwd_fp, rev_fp, args.output_dir,
                                      args.organism, args.pct, args.frac)
    save_summary(args.summary_file, config, summary_data)


def save_summary(f, config, data):
    result = {
        "program": "decontam",
        "version": __version__,
        "config": config,
        "data": data,
        }
    json.dump(result, f)


def make_index_main(argv=None):
    p = argparse.ArgumentParser()
    p.add_argument(
        "--config-file", required=True,
        type=argparse.FileType("r"),
        help="JSON configuration file")
    p.add_argument(
        "--organism", required=True,
        help="reference organism to filter from")
    args = p.parse_args(argv)
    config = get_config(args.config_file, args.organism)
    
    tool = FilteringTool(config)

    if not tool.index_exists():
        tool.make_index()



import pysam

def get_mapped_reads(fp, min_pct_id, min_len_frac):
    sam = pysam.AlignmentFile(fp)
    for read in sam:
        if read.is_unmapped:
            yield (read.query_name, read.is_read1, None)
        elif _get_frac(read) < min_len_frac:
            yield (read.query_name, read.is_read1, None)
        else:
            pct_id = _get_pct_identity(read)
            if pct_id < min_pct_id:
                yield (read.query_name, read.is_read1, None)
            else:
                ref_id = sam.getrname(read.reference_id)
                yield (read.query_name, read.is_read1, ref_id)


def _get_pct_identity(read):
    if read.has_tag("NM"):
        edit_dist = read.get_tag("NM")
    else:
        edit_dist = 0
    pct_mm = float(edit_dist) / read.alen
    return 1 - pct_mm 

def _get_frac(read):
    cigar = read.cigartuples
    clip = 0
    for pair in cigar:
        if pair[0] == 4 or pair[0] == 5:
            clip = clip + pair[1]
    frac = float(read.query_alignment_length)/(read.query_alignment_length + clip)
    return frac



import itertools
import os

class FastqSplitter(object):
    suffixes_host = {
        True: "_host",
        False: "",
        }
    suffixes_phix = {
        True: "_phix",
        False: "",
        }

    def __init__(self, input_fp, output_dir):
        self.output_dir = output_dir
        self.input_fp = input_fp
        input_filename = os.path.basename(input_fp)
        self.input_root, self.input_ext = os.path.splitext(input_filename)
        self._open_files = {}

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()

    def partition(self, annotations, organism):
        ids_to_annotation = dict(annotations)
        with open(self.input_fp) as f:
            for desc, seq, qual in parse_fastq(f):
                read_id = desc.split()[0]
                annotation = ids_to_annotation[read_id]
                self._write((desc, seq, qual), annotation, organism)

    def _write(self, read, annotation, organism):
        desc, seq, qual = read
        if organism == "host":
            suffix = self.suffixes_host[annotation]
        elif organism == "phix":
            suffix = self.suffixes_phix[annotation]
        output_filename = self.input_root + suffix + self.input_ext
        fp = os.path.join(self.output_dir, output_filename)
        if fp not in self._open_files:
            self._open_files[fp] = open(fp, "w")
        f = self._open_files[fp]
        write_fastq(f, desc, seq, qual)

    def close(self):
        for f in self._open_files.values():
            f.close()


def _grouper(iterable, n):
    "Collect data into fixed-length chunks or blocks"
    args = [iter(iterable)] * n
    return itertools.izip(*args)


def parse_fastq(f):
    """ parse original fastq file and write new fastq file the filtered non-host reads.
    """
    for desc, seq, _, qual in _grouper(f, 4):
        desc = desc.rstrip()[1:]
        seq = seq.rstrip()
        qual = qual.rstrip()
        yield desc, seq, qual


def write_fastq(out_fastq, desc, seq, qual):
    out_fastq.write("@" + desc + "\n")
    out_fastq.write(seq + "\n")
    out_fastq.write("+\n")
    out_fastq.write(qual + "\n")



import collections
import csv
import inspect
import os
import random
import re
import shutil
import subprocess
import tempfile
import utils


from decontamlib.fastq import FastqSplitter
from decontamlib.sam import get_mapped_reads


def FilteringTool(config):
    tool_cls = tools_available[config["method"]]
    # Proceed stepwise here to improve quality of error messages.
    tool_args = []
    for argname in tool_cls.get_argnames():
        arg = config[argname]
        tool_args.append(arg)
    return tool_cls(*tool_args)


class _FilteringTool(object):
    def __init__(self, index):
        self.index = index

    @classmethod
    def get_argnames(cls):
        return inspect.getargspec(cls.__init__)[0][1:]

    def decontaminate(self, fwd_fp, rev_fp, output_dir, organism, pct, frac):
        annotations = self.annotate(fwd_fp, rev_fp, pct, frac)
        with FastqSplitter(fwd_fp, output_dir) as s:
            s.partition(annotations, organism)
        with FastqSplitter(rev_fp, output_dir) as s:
            s.partition(annotations, organism)
        summary_data = summarize_annotations(annotations)
        return summary_data

    def annotate(self, fwd_fp, rev_fp, pct, frac):
        raise NotImplementedError()

    def _get_mapped_reads(self, filename, pct, frac):
        """Extracts set of qnames from SAM file."""
        mapped = set()
        for qname, is_read1, rname in get_mapped_reads(filename, pct, frac):
            if rname is not None:
                mapped.add(qname)
        return mapped

    def make_index(self):
        raise NotImplementedError()

    def index_exists(self):
        return True


class Bwa(_FilteringTool):
    def __init__(self, index, bwa_fp, num_threads):
        self.index = index
        self.bwa_fp = bwa_fp
        self.num_threads = num_threads
        
    def annotate(self, R1, R2, pct, frac):
        sam_file, stderr_file = self._run(R1, R2)
        mapped = self._get_mapped_reads(sam_file.name, pct, frac)
        ids = utils.parse_read_ids(R1)
        return [(id, True if id in mapped else False) for id in ids]

    def _command(self, fwd_fp, rev_fp):
        return [self.bwa_fp, "mem", "-M", "-t", str(self.num_threads), self.index, fwd_fp, rev_fp]

    def _run(self, R1, R2):
        stdout_file = tempfile.NamedTemporaryFile()
        stderr_file = tempfile.NamedTemporaryFile()
        command = self._command(R1, R2)
        subprocess.check_call(command, stdout=stdout_file, stderr=stderr_file)
        return (stdout_file, stderr_file)

    def make_index(self):
        cmd = [self.bwa_fp, "index", self.index]
        return subprocess.check_output(cmd, stderr=subprocess.STDOUT)

    def index_exists(self):
        return os.path.exists(self.index + ".amb")


class Bowtie(Bwa):
    def __init__(self, index, bowtie2_fp):
        self.index = index
        self.bowtie2_fp = bowtie2_fp

    def _command(self, fwd_fp, rev_fp):
        return [
            self.bowtie2_fp, "--local", "--very-sensitive-local",
            "-1", fwd_fp, "-2", rev_fp,
            "-x", self.index]

    def make_index(self):
        build_fp = self.bowtie2_fp + "-build"
        cmd = [build_fp, "-f", self.index, self.index]
        return subprocess.check_output(cmd, stderr=subprocess.STDOUT)

    def index_exists(self):
        index_fp = self.index + ".1.bt2"
        return os.path.exists(index_fp)


class Random_human(_FilteringTool):
    def __init__(self, percent_human):
        assert 0.0 <= percent_human <= 100.0
        self.fraction_human = percent_human / 100.0

    def annotate(self, R1, R2, pct, frac):
        ids = utils.parse_read_ids(R1)
        return [
            (id, True if random.random() <= self.fraction_human else False)
            for id in ids]


class None_human(_FilteringTool):
    def __init__(self):
        pass

    def annotate(self, R1, R2, pct, frac):
        ids = utils.parse_read_ids(R1)
        return [(id, False) for id in ids]


class All_human(_FilteringTool):
    def __init__(self):
        pass

    def annotate(self, R1, R2, pct, frac):
        ids = utils.parse_read_ids(R1)
        return [(id, True) for id in ids]


def summarize_annotations(annotations):
    return dict(collections.Counter(a for _, a in annotations))


tools_available = {
    "bwa": Bwa,
    "all_human": All_human,
    "no_human": None_human,
    "random_human": Random_human,
    "bowtie2": Bowtie,
}



from Bio import SeqIO
from pprint import pprint


def extract_column(filehandle, column_number, skip_lines=0, sep="\t"):
    """ gets unique values for a given column (column number starting 1) """
    ids = set()
    skip = 0
    for line in filehandle:
        if skip < skip_lines:
            skip +=1
            continue       
        line = line.strip()
        columns = line.split(sep)
        ids.add(columns[column_number - 1])
    return ids

def get_column(filehandle, column_number, sep="\t"):
    id = list()
    
    for line in filehandle:
        line = line.strip()
        columns = line.split(sep)              
        id.append(columns[column_number - 1])
    filehandle.seek(0)
    return id

def get_multiple_columns(filehandle, column_numbers, sep="\t"):
    data = []
    for line in filehandle:
        line = line.strip()
        columns = line.split(sep)
        data.append([ columns[number - 1] for number in column_numbers ])

    return data

def parse_read_ids(fastq_filename):
    """extract fastq ids.
    
    Returns:
        set of ids.
    """
    records = SeqIO.index(fastq_filename, "fastq")
    ids = set()
    for record in records:
        ids.add(record)
    return ids


def check_all_read_ids_are_consistent(id1, id2):
    """ check if ids are the same.
        
    Args:
        id1, id2 sets of ids
    Returns:
        True if sets are the same 
        False if setst are diffent
    """
    return set(id1) == set(id2)


def add_tool_sample(tool, sample, human_annotation):
    """ add tool and sample column for a given human annotation.
        
    Args:
        tool string with tool name
        sample stiring with sample name
        human_annotation list of tuples: (read_id, is_human). 

    Returns:
        list of tuples: each tuple is (tool, sample, read_id, is_human) value.
    """
    number_or_rows = len(human_annotation)
    tool_rows = [ tool ] * number_or_rows
    sample_rows = [ sample ] * number_or_rows
    (read_id, is_human) = zip(*human_annotation)
    return zip(tool_rows, sample_rows, read_id, is_human)
